@@ -1,11 +1,11 @@
 use strict;
 use warnings;
-package JSON::Schema::Tiny; # git description: v0.004-13-g4d862d6
+package JSON::Schema::Tiny; # git description: v0.005-10-ga1acb0b
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Validate data against a schema, minimally
 # KEYWORDS: JSON Schema data validation structure specification tiny
 
-our $VERSION = '0.005';
+our $VERSION = '0.006';
 
 use 5.016;  # for the unicode_strings feature
 no if "$]" >= 5.031009, feature => 'indirect';
@@ -16,7 +16,7 @@ use B;
 use Ref::Util 0.100 qw(is_plain_arrayref is_plain_hashref is_ref);
 use Mojo::URL;
 use Mojo::JSON::Pointer;
-use Carp 'croak';
+use Carp qw(croak carp);
 use Storable 'dclone';
 use JSON::MaybeXS 1.004001 'is_bool';
 use Feature::Compat::Try;
@@ -54,13 +54,13 @@ sub evaluate {
   my $state = {
     depth => 0,
     data_path => '',
-    traversed_schema_path => '',            # the accumulated path up to the last $ref traversal
-    canonical_schema_uri => Mojo::URL->new, # the canonical path of the last traversed $ref
-    schema_path => '',                      # the rest of the path, since the last traversed $ref
+    traversed_schema_path => '',          # the accumulated traversal path up to the last $ref traversal
+    initial_schema_uri => Mojo::URL->new, # the canonical URI as of the start or the last traversed $ref
+    schema_path => '',                    # the rest of the path, since the start or the last traversed $ref
     errors => [],
     seen => {},
     short_circuit => $BOOLEAN_RESULT || $SHORT_CIRCUIT,
-    root_schema => $schema,
+    root_schema => $schema,                 # so we can do $refs within the same document
   };
 
   my $valid;
@@ -88,6 +88,12 @@ sub evaluate {
 
 ######## NO PUBLIC INTERFACES FOLLOW THIS POINT ########
 
+# keyword => undef, or arrayref of alternatives
+my %removed_keywords = (
+  definitions => [ '$defs' ],
+  dependencies => [ qw(dependentSchemas dependentRequired) ],
+);
+
 sub _eval {
   croak '_eval called in void context' if not defined wantarray;
   croak 'insufficient arguments' if @_ < 3;
@@ -118,11 +124,13 @@ sub _eval {
 
   foreach my $keyword (
     # CORE KEYWORDS
-    qw($schema $ref $id $anchor $recursiveAnchor $vocabulary $comment $defs),
+    qw($id $schema $anchor $recursiveAnchor $ref $recursiveRef $vocabulary $comment $defs),
     # APPLICATOR KEYWORDS
     qw(allOf anyOf oneOf not if dependentSchemas
       items additionalItems contains
       properties patternProperties additionalProperties propertyNames),
+    # UNEVALUATED KEYWORDS
+    qw(unevaluatedItems unevaluatedProperties),
     # VALIDATOR KEYWORDS
     qw(type enum const
       multipleOf maximum exclusiveMaximum minimum exclusiveMinimum
@@ -146,15 +154,18 @@ sub _eval {
     last if not $valid and $state->{short_circuit};
   }
 
-  # UNSUPPORTED KEYWORDS
-  foreach my $keyword (
-    # CORE KEYWORDS
-    '$recursiveRef',
-    # APPLICATOR KEYWORDS
-    qw(unevaluatedItems unevaluatedProperties),
-  ) {
+  # check for previously-supported but now removed keywords
+  foreach my $keyword (keys %removed_keywords) {
     next if not exists $schema->{$keyword};
-    abort({ %$state, keyword => $keyword }, 'keyword not supported');
+    my $message ='no-longer-supported "'.$keyword.'" keyword present (at location "'
+      .canonical_schema_uri($state).'")';
+    if ($removed_keywords{$keyword}) {
+      my @list = map '"'.$_.'"', @{$removed_keywords{$keyword}};
+      @list = ((map $_.',', @list[0..$#list-1]), $list[-1]) if @list > 2;
+      splice(@list, -1, 0, 'or') if @list > 1;
+      $message .= ': this should be rewritten as '.join(' ', @list);
+    }
+    carp $message;
   }
 
   return $valid;
@@ -179,19 +190,57 @@ sub _eval_keyword_ref {
 
   return if not assert_keyword_type($state, $schema, 'string');
 
-  abort($state, 'only same-document JSON pointers are supported in $ref')
-    if $schema->{'$ref'} !~ m{^#(/(?:[^~]|~[01])*|)$};
+  my $uri = Mojo::URL->new($schema->{'$ref'})->to_abs($state->{initial_schema_uri});
+  abort($state, '$refs to anchors are not supported')
+    if ($uri->fragment//'') !~ m{^(/(?:[^~]|~[01])*|)$};
 
-  my $uri = Mojo::URL->new($schema->{'$ref'});
-  my $fragment = $uri->fragment;
+  # the base of the $ref uri must be the same as the base of the root schema
+  # unfortunately this means that many uses of $ref won't work, because we don't
+  # track the locations of $ids in this or other documents.
+  abort($state, 'only same-document, same-base JSON pointers are supported in $ref')
+    if $uri->clone->fragment(undef) ne Mojo::URL->new($state->{root_schema}{'$id'}//'');
 
-  my $subschema = Mojo::JSON::Pointer->new($state->{root_schema})->get($fragment);
+  my $subschema = Mojo::JSON::Pointer->new($state->{root_schema})->get($uri->fragment);
   abort($state, 'EXCEPTION: unable to find resource %s', $uri) if not defined $subschema;
 
   return _eval($data, $subschema,
     +{ %$state,
       traversed_schema_path => $state->{traversed_schema_path}.$state->{schema_path}.'/$ref',
-      canonical_schema_uri => $uri,
+      initial_schema_uri => $uri,
+      schema_path => '',
+    });
+}
+
+sub _eval_keyword_recursiveRef {
+  my ($data, $schema, $state) = @_;
+
+  return if not assert_keyword_type($state, $schema, 'string');
+
+  my $uri = Mojo::URL->new($schema->{'$recursiveRef'})->to_abs($state->{canonical_schema_uri});
+
+  abort($state, '$refs to anchors are not supported')
+    if ($uri->fragment//'') !~ m{^(/(?:[^~]|~[01])*|)$};
+
+  # the base of the $ref uri must be the same as the base of the root schema.
+  # unfortunately this means that nearly all usecases of $recursiveRef won't work, because we don't
+  # track the locations of $ids in this or other documents.
+  abort($state, 'only same-document, same-base JSON pointers are supported in $recursiveRef')
+    if $uri->clone->fragment(undef) ne Mojo::URL->new($state->{root_schema}{'$id'}//'');
+
+  my $subschema = Mojo::JSON::Pointer->new($state->{root_schema})->get($uri->fragment);
+  abort($state, 'EXCEPTION: unable to find resource %s', $uri) if not defined $subschema;
+
+  if (is_type('boolean', $subschema->{'$recursiveAnchor'}) and $subschema->{'$recursiveAnchor'}) {
+    $uri = Mojo::URL->new($schema->{'$recursiveRef'})
+      ->to_abs($state->{recursive_anchor_uri} // $state->{initial_schema_uri});
+    $subschema = Mojo::JSON::Pointer->new($state->{root_schema})->get($uri->fragment);
+    abort($state, 'EXCEPTION: unable to find resource %s', $uri) if not defined $subschema;
+  }
+
+  return _eval($data, $subschema,
+    +{ %$state,
+      traversed_schema_path => $state->{traversed_schema_path}.$state->{schema_path}.'/$recursiveRef',
+      initial_schema_uri => $uri,
       schema_path => '',
     });
 }
@@ -204,6 +253,12 @@ sub _eval_keyword_id {
   my $uri = Mojo::URL->new($schema->{'$id'});
   abort($state, '$id value should not equal "%s"', $uri) if $uri eq '' or $uri eq '#';
   abort($state, '$id value "%s" cannot have a non-empty fragment', $uri) if length $uri->fragment;
+
+  $uri->fragment(undef);
+  $state->{initial_schema_uri} = $uri->is_abs ? $uri : $uri->to_abs($state->{initial_schema_uri});
+  $state->{traversed_schema_path} = $state->{traversed_schema_path}.$state->{schema_path};
+  $state->{schema_path} = '';
+
   return 1;
 }
 
@@ -212,8 +267,8 @@ sub _eval_keyword_anchor {
 
   return if not assert_keyword_type($state, $schema, 'string');
 
-  return 1 if ($schema->{'$anchor'}//'') =~ /^[A-Za-z][A-Za-z0-9_:.-]*$/;
-  abort($state, '%s value does not match required syntax', $state->{keyword});
+  return 1 if $schema->{'$anchor'} =~ /^[A-Za-z][A-Za-z0-9_:.-]*$/;
+  abort($state, '$anchor value does not match required syntax');
 }
 
 sub _eval_keyword_recursiveAnchor {
@@ -227,6 +282,10 @@ sub _eval_keyword_recursiveAnchor {
   abort($state, '"$recursiveAnchor" keyword used without "$id"')
     if not exists $schema->{'$id'};
 
+  # record the canonical location of the current position, to be used against future resolution
+  # of a $recursiveRef uri -- as if it was the current location when we encounter a $ref.
+  $state->{recursive_anchor_uri} = canonical_schema_uri($state);
+
   return 1;
 }
 
@@ -247,6 +306,8 @@ sub _eval_keyword_vocabulary {
 
   abort($state, '$vocabulary can only appear at the document root')
     if length($state->{traversed_schema_path}.$state->{schema_path});
+
+  return 1;
 }
 
 sub _eval_keyword_comment {
@@ -853,6 +914,15 @@ sub _eval_keyword_propertyNames {
   return 1;
 }
 
+sub _eval_keyword_unevaluatedItems {
+  my ($data, $schema, $state) = @_;
+  abort($state, 'keyword not supported');
+}
+
+sub _eval_keyword_unevaluatedProperties {
+  my ($data, $schema, $state) = @_;
+  abort($state, 'keyword not supported');
+}
 
 # UTILITIES
 
@@ -984,7 +1054,7 @@ sub jsonp {
 sub canonical_schema_uri {
   my ($state, @extra_path) = @_;
 
-  my $uri = $state->{canonical_schema_uri}->clone;
+  my $uri = $state->{initial_schema_uri}->clone;
   $uri->fragment(($uri->fragment//'').jsonp($state->{schema_path}, @extra_path));
   $uri->fragment(undef) if not length($uri->fragment);
   $uri;
@@ -1001,7 +1071,7 @@ sub E {
     .jsonp($state->{schema_path}, $state->{keyword}, delete $state->{_schema_path_suffix});
 
   undef $uri if $uri eq '' and $keyword_location eq ''
-    or ($uri->fragment//'') eq $keyword_location;
+    or ($uri->fragment//'') eq $keyword_location and $uri->clone->fragment(undef) eq '';
 
   push @{$state->{errors}}, {
     instanceLocation => $state->{data_path},
@@ -1091,7 +1161,7 @@ JSON::Schema::Tiny - Validate data against a schema, minimally
 
 =head1 VERSION
 
-version 0.005
+version 0.006
 
 =head1 SYNOPSIS
 
